//---------------------------------------------------------------------------
#include <vcl.h>
#include <shlobj.h>
#include "ShellFunction.h"
//---------------------------------------------------------------------------
int __fastcall GetSystemImageList(bool large)
{
    SHFILEINFO info;

    return SHGetFileInfo("", 0, &info, sizeof(info), ((large) ? 0 : SHGFI_SMALLICON) | SHGFI_SYSICONINDEX);
}
//---------------------------------------------------------------------------
IShellFolder* __fastcall GetDesktopFolder()
{
    IShellFolder *folder;

    return (NOERROR == SHGetDesktopFolder(&folder)) ? folder : 0;
}
//---------------------------------------------------------------------------
IMalloc* __fastcall GetMalloc()
{
    IMalloc *malloc;

    return (NOERROR == SHGetMalloc(&malloc)) ? malloc : 0;
}
//---------------------------------------------------------------------------
ITEMIDLIST* __fastcall GetDesktopPidl(HWND hwnd)
{
    ITEMIDLIST *pidl;

    return (NOERROR == SHGetSpecialFolderLocation(hwnd, CSIDL_DESKTOP, &pidl)) ? pidl : 0;
}
//---------------------------------------------------------------------------
unsigned short __fastcall GetPidlLength(ITEMIDLIST *pidl)
{
    unsigned short length = 0;

    while(pidl->mkid.cb)
    {
        length += pidl->mkid.cb;
        (char*)pidl += pidl->mkid.cb;
    }

    return length;
}
//---------------------------------------------------------------------------
ITEMIDLIST* __fastcall CopyPidl(ITEMIDLIST *pidl, bool full)
{
    IMalloc *malloc = GetMalloc();
    ITEMIDLIST *newPidl = 0;

    if(malloc)
    {
        unsigned short length = (full) ? GetPidlLength(pidl) : pidl->mkid.cb;

        if(newPidl = (ITEMIDLIST*)malloc->Alloc(length + sizeof(unsigned short)))
        {
            memcpy(newPidl, pidl, length);
            *((unsigned short*)((char*)newPidl + length)) = 0;
        }

        malloc->Release();
    }

    return newPidl;
}
//---------------------------------------------------------------------------
ITEMIDLIST* __fastcall ConcatPidl(ITEMIDLIST *pidl1, ITEMIDLIST *pidl2)
{
    IMalloc *malloc = GetMalloc();
    ITEMIDLIST *newPidl = 0;

    if(malloc)
    {
        unsigned short length1 = GetPidlLength(pidl1);
        unsigned short length2 = GetPidlLength(pidl2);

        if(newPidl = (ITEMIDLIST*)malloc->Alloc(length1 + length2 + sizeof(unsigned short)))
        {
            memcpy(newPidl, pidl1, length1);
            memcpy((char*)newPidl + length1, pidl2, length2);
            *((unsigned short*)((char*)newPidl + length1 + length2)) = 0;
        }

        malloc->Release();
    }

    return newPidl;
}
//---------------------------------------------------------------------------
ITEMIDLIST* __fastcall ExtractPidl(ITEMIDLIST *pidl)
{
    while(((ITEMIDLIST*)((char*)pidl + pidl->mkid.cb))->mkid.cb)
        (char*)pidl += pidl->mkid.cb;

    return CopyPidl(pidl, false);
}
//---------------------------------------------------------------------------
IShellFolder* __fastcall GetShellFolder(ITEMIDLIST *pidl)
{
    IShellFolder *folder = 0;
    IMalloc *malloc = GetMalloc();

    if(malloc)
    {
        if(folder = GetDesktopFolder())
        {
            IShellFolder *parentFolder;
            ITEMIDLIST *currentPidl;

            while(pidl->mkid.cb)
            {
                parentFolder = folder;

                if(currentPidl = CopyPidl(pidl, false))
                {
                    parentFolder->BindToObject(currentPidl, 0, IID_IShellFolder, (void**)&folder);
                    malloc->Free(currentPidl);
                }
                else
                    folder = 0;

                parentFolder->Release();

                if(!folder)
                    break;

                (char*)pidl += pidl->mkid.cb;
            }
        }

        malloc->Release();
    }

    return folder;
}
//---------------------------------------------------------------------------
String __fastcall GetShellName(IShellFolder *folder, ITEMIDLIST *pidl)
{
    String name;
    IMalloc *malloc = GetMalloc();
    STRRET str;

    if(malloc)
    {
        if(NOERROR == folder->GetDisplayNameOf(pidl, SHGDN_NORMAL, &str))
            switch(str.uType)
            {
                case STRRET_CSTR:
                    name = str.cStr;
                    break;
                case STRRET_OFFSET:
                    name = (char*)pidl + str.uOffset;
                    break;
                case STRRET_WSTR:
                    name = str.pOleStr;
                    malloc->Free(str.pOleStr);
            }

        malloc->Release();
    }

    return name;
}
//---------------------------------------------------------------------------
int GetShellIcon(ITEMIDLIST *pidl, bool open, bool large)
{
    SHFILEINFO info;

    return (SHGetFileInfo((char*)pidl, 0, &info, sizeof(info), ((open) ? SHGFI_OPENICON : 0) | SHGFI_PIDL | ((large) ? 0 : SHGFI_SMALLICON) | SHGFI_SYSICONINDEX)) ? info.iIcon : -1;
}
//---------------------------------------------------------------------------
bool __fastcall FindPidlInit(HWND hwnd, IShellFolder *folder, unsigned long type, unsigned long attr, FindPidlRec &find)
{
    if(find.malloc = GetMalloc())
    {
        if(NOERROR == folder->EnumObjects(hwnd, type, &find.enumPidl))
        {
            find.folder = folder;
            find.attr = attr;
            return true;
        }

        find.malloc->Release();
    }

    return false;
}
//---------------------------------------------------------------------------
ITEMIDLIST* __fastcall FindPidl(FindPidlRec &find)
{
    unsigned long attr;
    unsigned long count;
    ITEMIDLIST *pidl;

    while(NOERROR == find.enumPidl->Next(1, (ITEMIDLIST**)&pidl, &count))
    {
        attr = find.attr;

        if(GetShellAttribute(find.folder, pidl, attr))
            return pidl;

        find.malloc->Free(pidl);
    }

    return 0;
}
//---------------------------------------------------------------------------
void __fastcall FindPidlClose(FindPidlRec &find)
{
    find.enumPidl->Release();
    find.malloc->Release();
}
//---------------------------------------------------------------------------
