//---------------------------------------------------------------------------
#include <vcl.h>
#pragma hdrstop

#include <shlobj.h>
#include "ShellFunction.h"
#include "ShellTreeView.h"
#pragma package(smart_init)
//---------------------------------------------------------------------------
static inline void ValidCtrCheck(TShellTreeView *)
{
    new TShellTreeView(NULL);
}
//---------------------------------------------------------------------------
__fastcall TShellTreeView::TShellTreeView(TComponent* Owner)
    : TWinControl(Owner)
{
    int sysImages = GetSystemImageList(false);

    if(!sysImages)
        throw Exception("Cannot retrieve the system image list handle");

    if(!(malloc = GetMalloc()))
        throw Exception("Cannot retrieve the IMalloc interface");

    tree = new TTreeView(this);
    images = new TImageList(this);
    images->Handle = sysImages;
    images->ShareImages = true;
    tree->Align = alClient;
    tree->Images = images;
    tree->ReadOnly = true;
    tree->ShowRoot = false;
    tree->SortType = stBoth;
    tree->OnChange = OnTreeViewChange;
    tree->OnExpanding = OnExpandTreeNode;
    tree->OnDeletion = OnDeleteTreeNode;
    tree->OnCompare = OnCompareTreeNode;
    tree->OnGetImageIndex = OnGetImageIndex;
    tree->OnGetSelectedIndex = OnGetSelectedIndex;
    Height = 250;
    Width = 200;
    TabStop = true;
}
//---------------------------------------------------------------------------
__fastcall TShellTreeView::~TShellTreeView()
{
    malloc->Release();
}
//---------------------------------------------------------------------------
namespace Shelltreeview
{
    void __fastcall PACKAGE Register()
    {
         TComponentClass classes[1] = {__classid(TShellTreeView)};
         RegisterComponents("BIU", classes, 0);
    }
}
//---------------------------------------------------------------------------
void __fastcall TShellTreeView::CreateWnd()
{
    TTreeNode *node;

    TWinControl::CreateWnd();
    tree->Parent = this;

    if(node = CreateRoot())
        node->Expand(false);
}
//---------------------------------------------------------------------------
void __fastcall TShellTreeView::OnTreeViewChange(System::TObject* Sender, TTreeNode* Node)
{
    if(FOnChange)
        FOnChange(this);
}
//---------------------------------------------------------------------------
void __fastcall TShellTreeView::OnExpandTreeNode(System::TObject* Sender, TTreeNode* Node, bool &AllowExpansion)
{
    if(!Node->getFirstChild())
    {
        ITEMIDLIST *parentPidl = GetFullPidl(Node);

        if(parentPidl)
        {
            IShellFolder *folder = GetShellFolder(parentPidl);

            if(folder)
            {
                FindPidlRec find;

                if(FindPidlInit(Handle, folder, SHCONTF_FOLDERS, SFGAO_FILESYSTEM | SFGAO_FILESYSANCESTOR, find))
                {
                    ITEMIDLIST *pidl;

                    while(pidl = FindPidl(find))
                        if(!CreateTreeNode(Node, folder, pidl))
                            malloc->Free(pidl);

                    FindPidlClose(find);
                }

                folder->Release();
            }

            malloc->Free(parentPidl);
        }
    }
}
//---------------------------------------------------------------------------
void __fastcall TShellTreeView::OnDeleteTreeNode(System::TObject* Sender, TTreeNode* Node)
{
    malloc->Free(Node->Data);
}
//---------------------------------------------------------------------------
void __fastcall TShellTreeView::OnCompareTreeNode(System::TObject* Sender, TTreeNode* Node1, TTreeNode* Node2, int Data, int &Compare)
{
    TTreeNode *node = Node1->Parent;

    if(node)
    {
        ITEMIDLIST *pidl = GetFullPidl(node);

        if(pidl)
        {
            IShellFolder *folder = GetShellFolder(pidl);

            if(folder)
            {
                Compare = ComparePidl(folder, (ITEMIDLIST*)Node1->Data, (ITEMIDLIST*)Node2->Data);
                folder->Release();
            }

            malloc->Free(pidl);
        }
    }
}
//---------------------------------------------------------------------------
void __fastcall TShellTreeView::OnGetImageIndex(System::TObject* Sender, TTreeNode* Node)
{
    if(Node->ImageIndex < 0)
    {
        ITEMIDLIST *pidl = GetFullPidl(Node);

        if(pidl)
        {
            Node->ImageIndex = GetShellIcon(pidl, false, false);
            malloc->Free(pidl);
        }
    }
}
//---------------------------------------------------------------------------
void __fastcall TShellTreeView::OnGetSelectedIndex(System::TObject* Sender, TTreeNode* Node)
{
    if(Node->SelectedIndex < 0)
    {
        ITEMIDLIST *pidl = GetFullPidl(Node);

        if(pidl)
        {
            Node->SelectedIndex = GetShellIcon(pidl, true, false);
            malloc->Free(pidl);
        }
    }
}
//---------------------------------------------------------------------------
ITEMIDLIST* __fastcall TShellTreeView::GetFullPidl(TTreeNode *node)
{
    ITEMIDLIST *pidl = CopyPidl((ITEMIDLIST*)node->Data, false);

    if(pidl)
    {
        ITEMIDLIST *childPidl;

        while(node = node->Parent)
        {
            childPidl = pidl;
            pidl = ConcatPidl((ITEMIDLIST*)node->Data, childPidl);
            malloc->Free(childPidl);

            if(!pidl)
                break;
        }
    }

    return pidl;
}
//---------------------------------------------------------------------------
TTreeNode* __fastcall TShellTreeView::CreateTreeNode(TTreeNode *node, IShellFolder *folder, ITEMIDLIST *pidl)
{
    if(node = tree->Items->AddChild(node, GetShellName(folder, pidl)))
    {
        unsigned long attr = SFGAO_HASSUBFOLDER;

        node->Data = pidl;
        node->ImageIndex = -1;
        node->SelectedIndex = -1;
        node->HasChildren = HasSubFolder(GetShellAttribute(folder, pidl, attr));
    }

    return node;
}
//---------------------------------------------------------------------------
TTreeNode* __fastcall TShellTreeView::CreateRoot()
{
    TTreeNode *node = 0;
    IShellFolder *folder = GetDesktopFolder();

    if(folder)
    {
        ITEMIDLIST *pidl = GetDesktopPidl(Handle);

        if(pidl)
            if(!(node = CreateTreeNode(0, folder, pidl)))
                malloc->Free(pidl);

        folder->Release();
    }

    return node;
}
//---------------------------------------------------------------------------
void __fastcall TShellTreeView::RefreshChildTreeNode(TTreeNode *node)
{
    ITEMIDLIST *parentPidl = GetFullPidl(node);

    if(parentPidl)
    {
        IShellFolder *folder = GetShellFolder(parentPidl);

        if(folder)
        {
            FindPidlRec find;

            if(FindPidlInit(Handle, folder, SHCONTF_FOLDERS, SFGAO_FILESYSTEM | SFGAO_FILESYSANCESTOR, find))
            {
                bool folderExist;
                TTreeNode *childNode;
                ITEMIDLIST *pidl;

                while(pidl = FindPidl(find))
                {
                    folderExist = false;
                    childNode = node->getFirstChild();

                    while(childNode)
                    {
                        if(!ComparePidl(folder, (ITEMIDLIST*)childNode->Data, pidl))
                        {
                            folderExist = true;
                            break;
                        }

                        childNode = childNode->getNextSibling();
                    }

                    if(!folderExist)
                        if(CreateTreeNode(node, folder, pidl))
                            continue;

                    malloc->Free(pidl);
                }

                FindPidlClose(find);
            }

            folder->Release();
        }

        malloc->Free(parentPidl);
    }
}
//---------------------------------------------------------------------------
void __fastcall TShellTreeView::RefreshTreeNode(TTreeNode *node)
{
    ITEMIDLIST *parentPidl = (node->Level) ? ((node->Parent) ? GetFullPidl(node->Parent) : 0) : GetFullPidl(node);

    if(parentPidl)
    {
        IShellFolder *folder = GetShellFolder(parentPidl);

        if(folder)
        {
            unsigned long attr = SFGAO_FILESYSTEM | SFGAO_FILESYSANCESTOR;
            ITEMIDLIST *pidl = (ITEMIDLIST*)node->Data;

            if(IsFileSystem(GetShellAttribute(folder, pidl, attr)))
            {
                ITEMIDLIST *fullPidl = ConcatPidl(parentPidl, pidl);

                attr = SFGAO_HASSUBFOLDER;
                node->Text = GetShellName(folder, pidl);

                if(fullPidl)
                {
                    node->ImageIndex = GetShellIcon(fullPidl, false, false);
                    node->SelectedIndex = GetShellIcon(fullPidl, true, false);
                    malloc->Free(fullPidl);
                }

                if(HasSubFolder(GetShellAttribute(folder, pidl, attr)))
                {
                    if(node->Expanded)
                        RefreshChildTreeNode(node);
                    else
                        {
                            node->DeleteChildren();
                            node->HasChildren = true;
                        }
                }
                else
                    node->DeleteChildren();
            }
            else
                node->Delete();

            folder->Release();
        }

        malloc->Free(parentPidl);
    }
}
//---------------------------------------------------------------------------
void __fastcall TShellTreeView::RefreshTree()
{
    TTreeNode *node = tree->Items->GetFirstNode();

    if(!node)
        if(!(node = CreateRoot()))
            return;

    do RefreshTreeNode(node);
    while(node = node->GetNextVisible());
}
//---------------------------------------------------------------------------
bool __fastcall TShellTreeView::GetHotTrack()
{
    return tree->HotTrack;
}
//---------------------------------------------------------------------------
void __fastcall TShellTreeView::SetHotTrack(bool hotTrack)
{
    tree->HotTrack = hotTrack;
}
//---------------------------------------------------------------------------
